"""
Holds information about the input domain
"""
from typing import List, Tuple

import numpy as np


def _pad_array(
        arr: np.ndarray,
        periodic: Tuple[bool, bool, bool],
        periodic_padding: int,
        extra_padding: int
        ) -> np.ndarray:
    """Add appropriate padding to the array"""

    # padding in periodic directions
    width = periodic_padding + extra_padding
    if width != 0:
        arr = np.pad(arr, [(width, width) if p else (0, 0) for p in periodic], 'wrap')

    # padding in non-periodic directions
    width = extra_padding
    if width != 0:
        arr = np.pad(arr, [(0, 0) if p else (width, width) for p in periodic], 'symmetric')

    return arr


class VOFDomain:
    """Holds the vof field and information about periodicity"""

    def __init__(self,
                 void_fraction: np.ndarray,
                 periodic: List[bool],
                 periodic_padding: int,
                 extra_padding: int
                 ):
        # Checks
        assert len(periodic) == void_fraction.ndim

        # Convert original VOF felid into a 3D field
        self.original_shape = void_fraction.shape
        match void_fraction.ndim:
            case 1: self.periodic = (False, periodic[0], False)
            case 2: self.periodic = (periodic[0], periodic[1], False)
            case 3: self.periodic = (periodic[0], periodic[1], periodic[2])
            case _:
                raise ValueError("Unexpected void_fraction.ndim: " + str(void_fraction.ndim))

        self._vof_storage = np.atleast_3d(void_fraction.copy())

        # Add padding
        self.periodic_padding = periodic_padding
        self.extra_padding = extra_padding
        self._vof_storage = _pad_array(self._vof_storage, self.periodic,
                                       periodic_padding=self.periodic_padding,
                                       extra_padding=self.extra_padding)

    def vof(self, padding: int = 0) -> np.ndarray:
        """
        Padding is `periodic_padding + padding` in periodic directions and `padding` in non-periodic directions
        """
        skip = self.extra_padding - padding
        assert skip >= 0, f"requested pad {padding} larger than domain's extra_padding {self.extra_padding}"

        if skip == 0:
            return self._vof_storage
        else:
            return self._vof_storage[skip:-(skip), skip:-(skip), skip:-(skip)]

    def convert_to_original_shape(self, arr: np.ndarray) -> np.ndarray:
        """
        Convert back to input shape of void_fraction
        """
        # remove padding in periodic directions
        if self.periodic_padding != 0:
            if self.periodic[0]:
                arr = arr[self.periodic_padding:-self.periodic_padding, :, :]
            if self.periodic[1]:
                arr = arr[:, self.periodic_padding:-self.periodic_padding, :]
            if self.periodic[2]:
                arr = arr[:, :, self.periodic_padding:-self.periodic_padding]

        return arr.reshape(self.original_shape)



import numpy as np
from blobid import get_labels

# import a VOF felid, for illustration we'll use a 2D slice
vof = 1.0 - np.load("tests/resources/fs_vof.npy")[:, 0, :]

# calculate labels
labels = get_labels(vof, periodic=[True, False])

# make the plots
import matplotlib.pyplot as plt
from matplotlib.colors import ListedColormap

# create a nice coloring for the labels
base = plt.colormaps["tab20"]
colors = [base(i % base.N) for i in np.unique(labels)]
colors[0] = (0, 0, 0, 1)  # unlabeled cells
labels_cmap = ListedColormap(colors)

vof_img = (1.0-vof.transpose()[::-1, :]).astype(np.float32)

fig, ax = plt.subplots(nrows=2)
ax[0].imshow(vof_img, vmin=0, vmax=1, cmap='binary')

ax[1].imshow(labels.transpose()[::-1, :], cmap=labels_cmap, rasterized=True)
ax[1].imshow(np.ones_like(vof_img), vmin=0, vmax=1, cmap='binary', alpha=vof_img)

for a in ax:
    a.set_xticks([])
    a.set_yticks([])
    a.set_aspect('equal')

# ensure deterministic SVG output
plt.rcParams['svg.hashsalt']="blobid-python"

plt.subplots_adjust(hspace=0)
plt.savefig("doc/example.svg", transparent=True, bbox_inches='tight', pad_inches=0, metadata={'Date': None})



import blobid as bi


def test_bench_base(benchmark, fs_vof_small):
    benchmark(bi.get_labels,
              void_fraction=fs_vof_small,
              periodic=[False, False, False]
              )


def test_bench_WY(benchmark, fs_vof_small):
    benchmark(bi.get_labels,
              void_fraction=fs_vof_small,
              periodic=[False, False, False],
              normals_method='WY'
              )


def test_bench_no_norm(benchmark, fs_vof_small):
    benchmark(bi.get_labels,
              void_fraction=fs_vof_small,
              periodic=[False, False, False],
              use_normals=False
              )


def test_bench_Chan(benchmark, fs_vof_small):
    benchmark(bi.get_labels,
              void_fraction=fs_vof_small,
              periodic=[False, False, False],
              use_normals=False,
              cutoff=0.5,
              cutoff_method='neighbors'
              )



import pytest
import numpy as np


@pytest.fixture
def fs_vof() -> np.ndarray:
    return 1.0 - np.load("tests/resources/fs_vof.npy")


@pytest.fixture
def fs_vof_small(fs_vof) -> np.ndarray:
    return fs_vof[
        :fs_vof.shape[0]//8,
        :fs_vof.shape[1]//8,
        fs_vof.shape[2]//3:-fs_vof.shape[2]//3]


@pytest.fixture
def fs_vof_medium(fs_vof) -> np.ndarray:
    return fs_vof[
        :fs_vof.shape[0]//2,
        :fs_vof.shape[1]//2,
        fs_vof.shape[2]//4:-fs_vof.shape[2]//4]


@pytest.fixture
def fs_vof_full(fs_vof) -> np.ndarray:
    return fs_vof



r"""
Implements piecewise linear interface calculation (PLIC), where the interface in each cell can be represented by
:math:`\vec{n}\cdot\vec{x}=\alpha`
"""

from .plic import get_normals, get_alpha, get_void_fraction, SUPPORTED_METHODS

from . import analytic_relations

__all__ = ['get_normals', 'get_alpha', 'get_void_fraction', 'SUPPORTED_METHODS', 'analytic_relations']



import hashlib

import numpy as np
import blobid as bi


def test_base(fs_vof):
    labels = bi.get_labels(fs_vof, periodic=[True, True, False],
                           use_normals=True)

    hash_object = hashlib.md5(labels.tobytes(order='C'))
    assert hash_object.hexdigest() == 'a8f17199d3bce93bba21ff1196cfe44c'


def test_WY_norm(fs_vof):
    labels = bi.get_labels(fs_vof.astype(np.float32), periodic=[True, True, False],
                           use_normals=True, normals_method='WY')

    hash_object = hashlib.md5(labels.tobytes(order='C'))
    assert hash_object.hexdigest() == 'e315fed4b5074c66a1a98cb65507e724'


def test_no_norm(fs_vof):
    labels = bi.get_labels(fs_vof, periodic=[True, True, False],
                           use_normals=False)

    hash_object = hashlib.md5(labels.tobytes(order='C'))
    assert hash_object.hexdigest() == '57eff32351abdf1640b77c08dfc5f9b0'


def test_cutoff(fs_vof):
    labels = bi.get_labels(fs_vof, periodic=[True, True, False],
                           use_normals=False,
                           cutoff=0.5,
                           cutoff_method='local')

    print(np.unique(labels).size)

    hash_object = hashlib.md5(labels.tobytes(order='C'))
    assert hash_object.hexdigest() == 'ee74eb206e28ec95c8466073b9f06dd6'


def test_Chan(fs_vof):
    labels = bi.get_labels(fs_vof, periodic=[True, True, False],
                           use_normals=False,
                           cutoff=0.5,
                           cutoff_method='neighbors')

    hash_object = hashlib.md5(labels.tobytes(order='C'))
    assert hash_object.hexdigest() == '7045d3492877ca9744e5064267bc20a5'



from typing import Tuple

import numpy as np

from .database import LabelDatabase
from .ccl import get_temporary_labels


def apply_ccl(
        is_object: np.ndarray,
        is_connected: np.ndarray,
        periodic: Tuple[bool, bool, bool],
        label_type
) -> np.ndarray:
    r"""
    Given connectedness, calculate unique labels for each connected region of object cells.

    Parameters
    ----------
    is_object : ndarray[ni, nj, nk]
        `is_object[i, j, k]` is true if cell `[i,j,k]` is an object cell
    is_connected : ndarray[ni, nj, nk, 3]
        `is_connected[i, j, k, d]` is true if cell `[i,j,k]` us connected to the neighbor in the negative d direction.
        For example, `is_connected[i, j, k, 0]` is true if cell `[i,j,k]` is connected to cell `[i-1,j,k]`
    periodic: (bool, bool, bool)
        If `periodic[d]`, then cells on each edge in direction `d` are considered the same cell.
        For example, cell `[0,j,k]` is the same as cell `[-1,j,k]` if `periodic[0]` is true.
    label_type : dtype
        Determines the integer data-type of `labels`

    Returns
    -------
        labels : ndarray[ni, nj, nk]
            An array of type `label_type` with the same shape as `is_object`.

    Notes
    -----
    The labeling algorithm used is based on work by He, Chao & Suzuki.[1]_


    .. [1] L. He, Y. Chao & K. Suzuki, "A Linear-Time Two-Scan Labeling Algorithm," IEEE International Conference
       on Image Processing, 2007. [10.1109/ICIP.2007.4379810](https://doi.org/10.1109/ICIP.2007.4379810)

    """
    # checks
    assert is_object.ndim == 3
    assert is_connected.ndim == 4
    assert np.all(is_object.shape == is_connected.shape[:-1])
    assert not np.any(is_connected[0, :, :, 0])
    assert not np.any(is_connected[:, 0, :, 1])
    assert not np.any(is_connected[:, :, 0, 2])

    # do initial labeling
    (labels, label_database) = get_temporary_labels(is_object, is_connected, label_type)

    # stitch together periodic boundaries
    label_database = _stitch_boundaries(labels, label_database, periodic)

    # do final labeling with sequential labels
    labels = label_database.get_sequential_lookup_table()[labels]

    return labels


def _stitch_boundaries(
        labels: np.ndarray,
        sets: LabelDatabase,
        periodic: Tuple[bool, bool, bool]
        ) -> LabelDatabase:
    """Stitch together periodic boundaries and remove padding"""

    if periodic[0]:
        for a, b in zip(labels[0, :, :].flat, labels[-2, :, :].flat):
            sets.merge(a, b)

    if periodic[1]:
        for a, b in zip(labels[:, 0, :].flat, labels[:, -2, :].flat):
            sets.merge(a, b)

    if periodic[2]:
        for a, b in zip(labels[:, :, 0].flat, labels[:, :, -2].flat):
            sets.merge(a, b)

    return sets



import numpy as np

import blobid as bi


def test_1d_gap():
    """Make sure that a region of f=0 causes separate blobs"""
    vof = np.array([1.0, 0.75, 0.0, 1.0, 1.0])

    lab_with_normals = bi.get_labels(vof, use_normals=True)
    assert np.all(lab_with_normals == [1, 1, 0, 2, 2])

    lab_with_normals = bi.get_labels(vof, use_normals=True, cutoff=0.8)
    assert np.all(lab_with_normals == [1, 0, 0, 2, 2])

    lab_without_normals = bi.get_labels(vof, use_normals=False)
    assert np.all(lab_without_normals == [1, 1, 0, 2, 2])

    lab_without_normals = bi.get_labels(vof, use_normals=False, cutoff=0.8)
    assert np.all(lab_without_normals == [1, 0, 0, 2, 2])


def test_1d_normals():
    """Test that normals correctly distinguish close blobs in 1D"""
    vof = np.array([1.0, 1.0, 0.25, 0.25, 1.0, 1.0])

    lab_with_normals = bi.get_labels(vof, use_normals=True)
    assert np.all(lab_with_normals == [1, 1, 1, 2, 2, 2])

    lab_without_normals = bi.get_labels(vof, use_normals=False)
    assert np.all(lab_without_normals == [1, 1, 1, 1, 1, 1])


def test_1d_cutoff():
    """Test that cutoff prevents bridging"""
    vof = np.array([0.1, 1.0, 1.0, 0.25, 0.25, 1.0, 1.0, 0.1])

    lab_without_normals = bi.get_labels(vof, use_normals=False, cutoff=0.8, cutoff_method='local')
    assert np.all(lab_without_normals == [0, 1, 1, 0, 0, 2, 2, 0])

    lab_without_normals = bi.get_labels(vof, use_normals=False, cutoff=0.8, cutoff_method='neighbors')
    assert np.all(lab_without_normals == [1, 1, 1, 1, 1, 1, 1, 1])

    vof = np.array([0.1, 0.1, 1.0, 1.0, 0.25, 0.25, 0.25, 1.0, 1.0, 0.1, 0.1])

    lab_without_normals = bi.get_labels(vof, use_normals=False, cutoff=0.8, cutoff_method='local')
    assert np.all(lab_without_normals == [0, 0, 1, 1, 0, 0, 0, 2, 2, 0, 0])

    lab_without_normals = bi.get_labels(vof, use_normals=False, cutoff=0.8, cutoff_method='neighbors')
    assert np.all(lab_without_normals == [0, 1, 1, 1, 1, 0, 2, 2, 2, 2, 0])


def test_neighbors_cutoff():
    """Test some special cases for neighbors-based cutoffs"""

    # make sure zeros aren't considered object cells
    vof = np.array([1, 0])
    lab = bi.get_labels(vof, use_normals=False, cutoff=0.5, cutoff_method='neighbors')
    assert np.all(lab == [1, 0])

    # make sure periodicity works
    vof = np.array([1, 0, 1, 0.25, 0.25])
    lab = bi.get_labels(vof, periodic=(True,), use_normals=False, cutoff=0.5, cutoff_method='neighbors')
    assert np.all(lab == [1, 0, 1, 1, 1])

    vof = np.array([0.25, 1, 0, 1, 0.25])
    lab = bi.get_labels(vof, periodic=(True,), use_normals=False, cutoff=0.5, cutoff_method='neighbors')
    assert np.all(lab == [1, 1, 0, 1, 1])

    vof = np.array([0.25, 0.25, 1, 0, 1])
    lab = bi.get_labels(vof, periodic=(True,), use_normals=False, cutoff=0.5, cutoff_method='neighbors')
    assert np.all(lab == [1, 1, 1, 0, 1])


def test_1d_periodic():
    """Test that a single periodic boundary works in 1D"""
    vof = np.array([1, 0, 1, 0, 1])

    # without periodicity (default behavior)
    labels = bi.get_labels(vof, use_normals=False)
    assert np.all(labels == [1, 0, 2, 0, 3])

    # with periodicity
    labels = bi.get_labels(vof, periodic=(True,), use_normals=False)
    assert np.all(labels == [1, 0, 2, 0, 1])


def test_2d_periodic():
    """Test that periodic boundaries work"""
    vof = np.array(
        [
            [0, 0, 1, 0, 0],
            [1, 0, 0, 0, 1],
            [0, 0, 1, 0, 0],
        ]
    )

    # without periodicity (default behavior)
    labels = bi.get_labels(vof, use_normals=False)
    assert np.all(labels == np.array(
        [
            [0, 0, 1, 0, 0],
            [2, 0, 0, 0, 3],
            [0, 0, 4, 0, 0],
        ]
    ))

    # periodic in x
    labels = bi.get_labels(
        vof, periodic=[True, False], use_normals=False
    )
    assert np.all(labels == np.array(
        [
            [0, 0, 1, 0, 0],
            [2, 0, 0, 0, 3],
            [0, 0, 1, 0, 0],
        ]
    ))

    # periodic in y
    labels = bi.get_labels(
        vof, periodic=[False, True], use_normals=False
    )
    assert np.all(labels == np.array(
        [
            [0, 0, 1, 0, 0],
            [2, 0, 0, 0, 2],
            [0, 0, 3, 0, 0],
        ]
    ))

    # periodic in x and y
    labels = bi.get_labels(
        vof, periodic=[True, True], use_normals=False
    )
    assert np.all(labels == np.array(
        [
            [0, 0, 1, 0, 0],
            [2, 0, 0, 0, 2],
            [0, 0, 1, 0, 0],
        ]
    ))


def test_u_shape():
    """Test of a more complex 2D shape"""
    vof = np.array([[1, 0, 1], [1, 0, 1], [1, 1, 1]])

    for i in range(4):
        lab = bi.get_labels(np.rot90(vof, i), use_normals=True)
        assert np.all(np.rot90(lab, -i) == vof)


def test_odd_label_type():
    vof = np.array([1.0, 0.0, 1.0])

    for label_type in [np.uint32, np.int16]:
        lab = bi.get_labels(vof, use_normals=False, label_type=label_type)

        assert np.all(lab == [1, 0, 2])
        assert lab.dtype == label_type



"""
Tools for running connected component labeling
"""

import numpy as np
from .._numba_support import njit
from .database import LabelDatabase, _DatabaseStorage, _merge


def get_temporary_labels(
        is_object: np.ndarray,
        is_connected: np.ndarray,
        label_type
        ) -> tuple[np.ndarray, LabelDatabase]:
    """Perform initial labeling and build the set of label connections"""

    # head cells that have no connections in the negative direction
    is_head = np.logical_and(
        is_object,
        np.logical_not(np.logical_or.reduce(is_connected, axis=3))
    )
    # all other object cells are "tail" cells
    is_tail = np.logical_and(
        is_object,
        np.logical_not(is_head)
    )

    # initialize data types
    label_database = LabelDatabase(label_count=np.count_nonzero(is_head), label_type=label_type)
    labels = np.zeros_like(is_object, dtype=label_database.label_type)

    # First pass: assigning each head cell a label
    labels[is_head] = label_database.get_labels()

    # Second pass : assign labels to "tail" cells and deal with merging
    _tail_pass(labels, label_database.data, is_tail, is_connected)

    return (labels, label_database)


@njit
def _tail_pass(
        labels: np.ndarray,
        data: _DatabaseStorage,
        is_tail: np.ndarray,
        is_connected: np.ndarray
        ):
    """Apply labels to tail cells and deal with merging"""

    neighbor_offsets = [(-1, 0, 0), (0, -1, 0), (0, 0, -1)]
    dummy_label = labels.dtype.type(0)

    for i, j, k in np.ndindex(labels.shape):
        if not is_tail[i, j, k]:
            continue

        new_label = dummy_label

        # look to the connected neighboring cells in the negative directions
        for d, offset in enumerate(neighbor_offsets):
            if is_connected[i, j, k, d]:
                neighbor_label = labels[i+offset[0], j+offset[1], k+offset[2]]

                if new_label == dummy_label:
                    # copy the neighbors label
                    new_label = neighbor_label
                elif new_label != neighbor_label:
                    # merge the two labels
                    new_label = _merge(data, new_label, neighbor_label)

        # set the label
        labels[i, j, k] = new_label



import pytest
import numpy as np

from blobid import plic

from blobid._domain import _pad_array
from blobid.labeling.database import LabelDatabase
from blobid.labeling.labeling import _stitch_boundaries


@pytest.fixture
def vof_3d() -> np.ndarray:
    """Create a dummy VOF array"""
    return np.random.rand(8*7*4).reshape((8, 7, 4))


def check_periodicity(arr, periodicity, depth):
    if periodicity[0] and depth != 0:
        assert np.all(arr[:depth, :, :] == arr[-2*depth:-depth, :, :])
        assert np.all(arr[depth:2*depth, :, :] == arr[-depth:, :, :])

    if periodicity[1] and depth != 0:
        assert np.all(arr[:, :depth, :] == arr[:, -2*depth:-depth, :])
        assert np.all(arr[:, depth:2*depth, :] == arr[:, -depth:, :])

    if periodicity[2] and depth != 0:
        assert np.all(arr[:, :, :depth] == arr[:, :, -2*depth:-depth])
        assert np.all(arr[:, :, depth:2*depth] == arr[:, :, -depth:])


def check_symmetry(arr, symmetry, depth):
    if symmetry[0] and depth != 0:
        assert np.all(arr[:depth, :, :] == np.flip(arr[depth:2*depth, :, :], axis=0))
        assert np.all(arr[-depth:, :, :] == np.flip(arr[-2*depth:-depth, :, :], axis=0))

    if symmetry[1] and depth != 0:
        assert np.all(arr[:, :depth, :] == np.flip(arr[:, depth:2*depth, :], axis=1))
        assert np.all(arr[:, -depth:, :] == np.flip(arr[:, -2*depth:-depth, :], axis=1))

    if symmetry[2] and depth != 0:
        assert np.all(arr[:, :, :depth] == np.flip(arr[:, :, depth:2*depth], axis=2))
        assert np.all(arr[:, :, -depth:] == np.flip(arr[:, :, -2*depth:-depth], axis=2))


def test_pad_array(vof_3d):
    for periodic_size in range(3):
        for extra in range(3):
            for n in range(8):
                per = [bool(n % 2), bool((n//2) % 2), bool((n//4) % 2)]

                arr = _pad_array(vof_3d.copy(), per, periodic_size, extra)

                # make sure dimensions are right
                for d in range(3):
                    assert arr.shape[d] == vof_3d.shape[d]+2*per[d]*periodic_size+2*extra

                # make sure nothing else has changed
                unchanged_range = arr[
                    (per[0]*periodic_size+extra):arr.shape[0]-(per[0]*periodic_size+extra),
                    (per[1]*periodic_size+extra):arr.shape[1]-(per[1]*periodic_size+extra),
                    (per[2]*periodic_size+extra):arr.shape[2]-(per[2]*periodic_size+extra)
                ]
                assert np.all(unchanged_range == vof_3d)

                # check edges
                check_periodicity(arr, per, extra+periodic_size)
                if extra != 0:
                    check_symmetry(arr, [not p for p in per], extra)


def test_stitch_boundaries():  # noqa: C901
    def create_label_field(n):
        sets = LabelDatabase(n[0]*n[1]*n[2])
        labels = np.zeros(n[0]*n[1]*n[2], dtype=sets.label_type)

        labels.flat = sets.get_labels()

        return (labels.reshape(n), sets)

    def check_periodicity(labels, sets, periodicity):
        if periodicity[0]:
            for a, b in zip(labels[1, :, :].flat, labels[-2, :, :].flat):
                assert sets.root(a) == sets.root(b)
        else:
            for a, b in zip(labels[0, :, :].flat, labels[-1, :, :].flat):
                assert sets.root(a) != sets.root(b)

        if periodicity[1]:
            for a, b in zip(labels[:, 1, :].flat, labels[:, -2, :].flat):
                assert sets.root(a) == sets.root(b)
        else:
            for a, b in zip(labels[:, 0, :].flat, labels[:, -1, :].flat):
                assert sets.root(a) != sets.root(b)

        if periodicity[2]:
            for a, b in zip(labels[:, :, 1].flat, labels[:, :, -2].flat):
                assert sets.root(a) == sets.root(b)
        else:
            for a, b in zip(labels[:, :, 0].flat, labels[:, :, -1].flat):
                assert sets.root(a) != sets.root(b)

    for n in range(8):
        per = [bool(n % 2), bool((n//2) % 2), bool((n//4) % 2)]

        (labels, sets) = create_label_field((5+2*per[0], 3+2*per[1], 4+2*per[2]))

        # link
        if per[0]:
            for a, b in zip(labels[0, :, :].flat, labels[1, :, :].flat):
                sets.merge(a, b)
            for a, b in zip(labels[-1, :, :].flat, labels[-2, :, :].flat):
                sets.merge(a, b)

        if per[1]:
            for a, b in zip(labels[:, 0, :].flat, labels[:, 1, :].flat):
                sets.merge(a, b)
            for a, b in zip(labels[:, -1, :].flat, labels[:, -2, :].flat):
                sets.merge(a, b)

        if per[2]:
            for a, b in zip(labels[:, :, 0].flat, labels[:, :, 1].flat):
                sets.merge(a, b)
            for a, b in zip(labels[:, :, -1].flat, labels[:, :, -2].flat):
                sets.merge(a, b)

        sets = _stitch_boundaries(labels, sets, per)

        # make sure the size is right
        assert labels.shape[0] == 5 + 2 * per[0]
        assert labels.shape[1] == 3 + 2 * per[1]
        assert labels.shape[2] == 4 + 2 * per[2]

        check_periodicity(labels, sets, per)


def test_normal_calculation_at_boundaries(vof_3d):
    for n in range(8):
        per = [bool(n % 2), bool((n//2) % 2), bool((n//4) % 2)]
        f = _pad_array(vof_3d.copy(), per, 1, 1)

        # normals should be the same on periodic sides
        n = plic.get_normals(f, normals_method='CD')
        check_periodicity(np.moveaxis(n, 0, 3), per, 1)

        n = plic.get_normals(f, normals_method='WY')
        check_periodicity(np.moveaxis(n, 0, 3), per, 1)



"""
Main solver routine for blobid
"""
from typing import List

import numpy as np

from ._domain import VOFDomain

from . import labeling
from . import plic


def get_labels(
        void_fraction: np.ndarray,
        periodic: List[bool] | None = None,
        cutoff: float = 0.0,
        cutoff_method: str = 'local',
        use_normals: bool = True,
        normals_method: str = 'CD',
        label_type=None
        ) -> np.ndarray:
    r"""
    Calculate unique labels for each connected region of `void_fraction` (blob) on a Cartesian grid.

    Parameters
    ----------
    void_fraction : ndarray
        The void fractions for each grid cell. The number of dimensions specifies the dimensionality of the domain.
        This function can handel 1-D, 2-D, or 3-D domains.
    periodic : 1-D array of bools, optional
        Sets the periodicity of the domain. Default is no direction is periodic.
    cutoff : float, optional
        The minimum value of `void_fraction` for a cell to be in region. Defaults to 0.
    cutoff_method : {'local', 'neighbors'}, optional
        How `cutoff` is applied (see Notes). Defaults to 'local'.
    use_normals : bool, optional
        Whether to use interface normals to determine connectedness. Defaults to True.
    normals_method: {'CD', 'WY'}, optional
        Passed to [plic.get_normals()](blobid/plic.html#get_normals) for calculating interface
        normals.
    label_type : dtype, optional
        Determines the integer data-type of `labels`. Defaults to `numpy.uint32`

    Returns
    -------
        labels : ndarray
            An array of type `label_type` with the same shape as `void_fraction`.

    Raises
    ------
        ValueError
            If the dimensions of `void_fraction` is not 1, 2, or 3.

        OverflowError
            If the number of labels exceeds the capacity of `label_type`.
            Note that the number of intermediate labels can exceed the largest label returned in `labels`

    Notes
    =====

    Object Criteria
    ---------------
    - If `cutoff_method` is `'local'`, a grid cell at :math:`{ijk}` will be an object cell if it's `void_fraction`
    :math:`f_{ijk}` is greater than `cutoff` :math:`f_c`.
    .. math::
        f_{ijk}>f_c

    - If `cutoff_method` is `'neighbors'`, a grid cell at :math:`{ijk}` will be an object cell if it's `void_fraction`
    :math:`f_{ijk}` is non-zero and it or any of its neighbors (6 in 3-D) have a `void_fraction` greater than
    `cutoff`.[1]_
    .. math::
        \Big[f_{ijk}>0\Big] \land
        \Big[ [f_{ijk}>f_{c}] \lor [f_{i+1jk}>f_{c}] \lor [f_{i-1jk}>f_{c}] \lor [f_{ij+1k}>f_{c}] \lor \cdots \Big]

    All object cells will be assigned to a blob.

    Connectivity Criteria
    ---------------------
    In 3-D, each cell has 6 neighbor cells it can be connected to.
    The first requirement for two neighbor cells to be connected is that they are both object cells.
    In addition (unless `use_normals` is False), the interface normals of each cell must not be opposed.[2]_
    Interface normals are calculated using the method set by `normals_method`
    (see [reconstruction.get_normals()](blobid/reconstruction.html#get_normals)).

    Labeling
    --------
    Based on this object and connectivity criteria, labels are assigned to each blob (defined as a connected set
    of object cells) with a two-scan labeling algorithm.[3]_

    Examples
    ========
    A simple 1-D example with four cells:

    >>> import numpy as np
    >>> from blobid import get_labels
    >>> f = np.array([1.0,0.25,0.25,1.0])
    >>> labels = get_labels(f)
    >>> labels
    array([1, 1, 2, 2], dtype=uint32)

    The interface normals in the second and third cell are opposed so they are not connected, and two distinct blobs
    are identified.
    If we set `use_normals` to false, only one blob is identified.

    >>> labels = get_labels(f, use_normals=False)
    >>> labels
    array([1, 1, 1, 1], dtype=uint32)

    We also see only one blob is identified if the domain is periodic because the first and last cells are connected.
    >>> labels = get_labels(f, periodic=(True,))
    >>> labels
    array([1, 1, 1, 1], dtype=uint32)

    If we only want to consider areas where the void fraction is larger than 0.5, we can do this with `cutoff`
    >>> labels = get_labels(f, cutoff=0.5)
    >>> labels
    array([1, 0, 0, 2], dtype=uint32)

    Note that a label of 0 indicates a cell was not assigned to a blob.

    References
    ----------
    .. [1] W. H. R. Chan, M. S. Dodd, P. L. Johnson, P. Moin, "Identifying and tracking bubbles and drops in
       simulations: A toolbox for obtaining sizes, lineages, and breakup and coalescence statistics," Journal of
       Computational Physics, vol. 432, pp. 110156, 2021.
       [10.1016/j.jcp.2021.110156](https://doi.org/10.1016/j.jcp.2021.110156)

    .. [2] K. Hendrickson, G. D. Weymouth & D. K.-P. Yue, "Informed component label algorithm for robust identification
       of connected components with volume-of-fluid method," Computers & Fluids, vol. 197, pp. 104373, 2020.
       [10.1016/j.compfluid.2019.104373](https://doi.org/10.1016/j.compfluid.2019.104373)

    """

    # Defaults
    if label_type is None:
        label_type = np.uint32

    # Setup to domain
    domain = VOFDomain(void_fraction=void_fraction,
                       periodic=periodic if periodic is not None else [False] * void_fraction.ndim,
                       periodic_padding=1,
                       extra_padding=1 if (use_normals or (cutoff_method == 'neighbors')) else 0)

    # calculate object cells
    is_object = _calc_object_cells(domain, cutoff, cutoff_method)

    # calculate connectivity
    if use_normals:
        normals = plic.get_normals(domain.vof(padding=1), normals_method)
    else:
        normals = None

    is_connected = _calc_connections(is_object, norm=normals)

    # do the labeling
    labels = labeling.apply_ccl(
        is_object=is_object,
        is_connected=is_connected,
        periodic=domain.periodic,
        label_type=label_type
        )

    # reshape to original dimensions (removes padding in periodic directions)
    return domain.convert_to_original_shape(labels)


def _calc_object_cells(domain: VOFDomain, cutoff: float, cutoff_method: str) -> np.ndarray:
    """returns an array that is true if cell is an object cell"""

    match cutoff_method:
        case 'local':
            return domain.vof() > cutoff
        case 'neighbors':
            large = domain.vof(padding=1) > cutoff

            large_neighbor = np.logical_or.reduce([
                large[1:-1, 1:-1, 1:-1],  # center
                large[2:,   1:-1, 1:-1],  # i+1
                large[:-2,  1:-1, 1:-1],  # i-1
                large[1:-1, 2:,   1:-1],  # j+1
                large[1:-1, :-2,  1:-1],  # j-1
                large[1:-1, 1:-1, 2:],    # k+1
                large[1:-1, 1:-1, :-2],   # k-1
            ])

            return np.logical_and(large_neighbor, domain.vof() > 0)

        case _:
            raise ValueError(f"cutoff_method '{cutoff_method}' is not supported")


def _calc_connections(is_object: np.ndarray, norm: np.ndarray | None) -> np.ndarray:
    """arr[i,j,k,d] is true if cell [i,j,k] is connected to the neighbor in the negative d direction"""

    # by default, false
    arr = np.zeros((3, is_object.shape[0], is_object.shape[1], is_object.shape[2]), dtype=bool)

    # cell and neighbor must be object cells
    arr[0, 1:, :, :] = is_object[:-1, :, :] & is_object[1:, :, :]
    arr[1, :, 1:, :] = is_object[:, :-1, :] & is_object[:, 1:, :]
    arr[2, :, :, 1:] = is_object[:, :, :-1] & is_object[:, :, 1:]

    # If using normals,
    # not connected if adjacent normals look like | --> | <-- |
    if norm is not None:
        arr[0, 1:, :, :] &= np.logical_or(norm[0, :-1, :, :] <= 0, norm[0, 1:, :, :] >= 0)
        arr[1, :, 1:, :] &= np.logical_or(norm[1, :, :-1, :] <= 0, norm[1, :, 1:, :] >= 0)
        arr[2, :, :, 1:] &= np.logical_or(norm[2, :, :, :-1] <= 0, norm[2, :, :, 1:] >= 0)

    return np.moveaxis(arr, 0, 3)



"""
Implementation of simple central differencing for interface reconstruction
"""
import numpy as np
import numpy.typing as npt


def get_normals(f: np.ndarray, normals_dtype: npt.DTypeLike) -> np.ndarray:
    """
    Calculate the sign of the interface normal using central difference
    """

    n = np.empty((3, f.shape[0]-2, f.shape[1]-2, f.shape[2]-2), dtype=normals_dtype)

    n[0] = (f[:-2, 1:-1, 1:-1] - f[2:, 1:-1, 1:-1])/2.0
    n[1] = (f[1:-1, :-2, 1:-1] - f[1:-1, 2:, 1:-1])/2.0
    n[2] = (f[1:-1, 1:-1, :-2] - f[1:-1, 1:-1, 2:])/2.0

    return n



"""Setup Numba if available"""

numba_availible = True
try:
    from numba import njit
except ImportError:
    numba_availible = False

    # create a dummy njit decorator
    def njit(func): return func



import hashlib

import pytest

import numpy as np
import blobid.plic as plic


def test_undefined_normals():
    """Tests where normal should be all zeros (undefined)"""

    def assert_undefined(norm):
        n = norm.squeeze()
        assert n[0] == 0
        assert n[1] == 0
        assert n[2] == 0

    # all zeros
    f = np.zeros((3, 3, 3))
    for method in plic.SUPPORTED_METHODS:
        assert_undefined(plic.get_normals(f, method))

    # all ones
    f = np.ones((3, 3, 3))
    for method in plic.SUPPORTED_METHODS:
        assert_undefined(plic.get_normals(f, method))

    # any value in a range between 0 and 1
    for _ in range(1000):
        f = np.ones((3, 3, 3)) * np.random.rand()
        for method in plic.SUPPORTED_METHODS:
            assert_undefined(plic.get_normals(f, method))


def test_end_to_end(fs_vof):
    expected_result = [
        ['CD', '599820a35d442746d95e8cd2f06fbc61'],
        ['WY', '2aeb743728db4ecefda7ce7527d836b4']
    ]

    for method, normals_signature in expected_result:
        normals = plic.get_normals(fs_vof.astype(np.float32), method)

        hash_object = hashlib.md5(normals.tobytes(order='C'))
        assert hash_object.hexdigest() == normals_signature


def test_round_trip(fs_vof_medium):
    normals = plic.get_normals(fs_vof_medium.astype(np.float64), 'CD')
    undefined_normal = np.all(normals == 0, axis=0)

    void_fraction = fs_vof_medium[1:-1, 1:-1, 1:-1].astype(np.float64)

    alpha = plic.get_alpha(void_fraction, normals)
    void_fraction_test = plic.get_void_fraction(alpha, normals)

    with np.nditer([void_fraction, void_fraction_test, undefined_normal]) as itr:
        for f, f_test, undef in itr:
            if np.isnan(f_test):
                assert undef or f == 0 or f == 1
            else:
                assert f_test == pytest.approx(f)


def test_bench_get_alpha(benchmark, fs_vof_small):
    normals = plic.get_normals(fs_vof_small.astype(np.float32), 'WY')
    void_fraction = fs_vof_small[1:-1, 1:-1, 1:-1].astype(np.float32)

    benchmark(plic.get_alpha,
              void_fraction=void_fraction,
              normals=normals)


def test_bench_get_void_fraction(benchmark, fs_vof_small):
    normals = plic.get_normals(fs_vof_small.astype(np.float32), 'WY')
    void_fraction = fs_vof_small[1:-1, 1:-1, 1:-1].astype(np.float32)
    alpha = plic.get_alpha(void_fraction, normals)

    benchmark(plic.get_void_fraction,
              alpha=alpha,
              normals=normals)



import warnings

import pytest
import numpy as np
import blobid.plic.nivira as nivira

from blobid._numba_support import numba_availible


def test_WY_Fig2():
    """Figure 2 from Weymouth and Yue"""

    f = np.empty((3, 3, 3))
    for k in range(3):
        f[:, :, k] = np.array([[1.0, 1.0, 0.95],
                               [1.0, 0.9, 0.3],
                               [0.7, 0.15, 0.0]])

    assert nivira._get_dominant_direction(f)[0, 0, 0] == 0

    n = nivira.get_normals(f, f.dtype).squeeze()
    assert n[0] == pytest.approx((1.0-0.15)/2.0)
    assert n[1] == pytest.approx(2.05-1.25)
    assert n[2] == 0


def test_vof_type():
    f = np.random.rand(3, 3, 3)

    # float16 should give a warning
    if numba_availible:
        with pytest.warns():
            nivira.get_normals(f.astype(np.float16), np.float32)

        with pytest.warns():
            nivira.get_normals(f.astype(np.float32), np.float16)

    # float32 and float64 should work without warning
    with warnings.catch_warnings():
        warnings.simplefilter("error")
        nivira.get_normals(f.astype(np.float32), np.float32)
        nivira.get_normals(f.astype(np.float64), np.float32)



import pytest as pt
import numpy as np

from blobid.labeling.database import LabelDatabase


def test_label_sets():
    s = LabelDatabase(7)

    # create 7 sets
    for i in range(7):
        assert i+1 == s.get_labels()[i]

    # check they return the right labels
    assert 1 == s.root(1)
    assert 2 == s.root(2)
    assert 3 == s.root(3)
    assert 4 == s.root(4)
    assert 5 == s.root(5)
    assert 6 == s.root(6)

    # merge two
    assert s.merge(1, 6) == 1
    assert 1 == s.root(1)
    assert 2 == s.root(2)
    assert 3 == s.root(3)
    assert 4 == s.root(4)
    assert 5 == s.root(5)
    assert 1 == s.root(6)

    # merging the same should not break anything
    assert s.merge(3, 3) == 3
    assert 1 == s.root(1)
    assert 2 == s.root(2)
    assert 3 == s.root(3)
    assert 4 == s.root(4)
    assert 5 == s.root(5)
    assert 1 == s.root(6)

    # merge two others
    assert s.merge(4, 5) == 4
    assert 1 == s.root(1)
    assert 2 == s.root(2)
    assert 3 == s.root(3)
    assert 4 == s.root(4)
    assert 4 == s.root(5)
    assert 1 == s.root(6)

    # merge group
    assert s.merge(6, 4) == 1
    assert 1 == s.root(1)
    assert 2 == s.root(2)
    assert 3 == s.root(3)
    assert 1 == s.root(4)
    assert 1 == s.root(5)
    assert 1 == s.root(6)

    # merge within same group should have no effect
    assert s.merge(5, 1) == 1
    assert 1 == s.root(1)
    assert 2 == s.root(2)
    assert 3 == s.root(3)
    assert 1 == s.root(4)
    assert 1 == s.root(5)
    assert 1 == s.root(6)

    # merge everything
    assert s.merge(2, 3) == 2
    assert s.merge(3, 1) == 1
    assert 1 == s.root(1)
    assert 1 == s.root(2)
    assert 1 == s.root(3)
    assert 1 == s.root(4)
    assert 1 == s.root(5)
    assert 1 == s.root(6)

    assert s.merge(2, 2) == 1


def test_label_sets_failures():
    # cant create labels with a non-integer type
    with pt.raises(AssertionError):
        s = LabelDatabase(20, float)  # noqa: F841

    # test running out of space
    s = LabelDatabase(255, np.uint8)  # noqa: F841

    with pt.raises(OverflowError):
        s = LabelDatabase(256, np.uint8)  # noqa: F841


def test_sequential_labels():
    s = LabelDatabase(20)

    # merge pairs
    for d in range(10):
        assert s.merge(2*d+1, 2*d+2) == 2*d+1

    lookup = s.get_sequential_lookup_table()

    assert lookup[0] == 0
    assert np.unique(lookup[1:]).size == 10

    for lab in range(1, 21):
        assert s.root(lab) == lab if lab % 2 else lab-1
        assert lookup[lab] == (lab-1)//2 + 1


def test_label_sets_types():
    for label_type in [np.uint32, int, np.int16]:
        s = LabelDatabase(20, label_type)

        assert s.label_type == label_type

        for i in range(20):
            new_label = s.get_labels()[i]
            assert i+1 == new_label
            assert np.dtype(label_type) == type(new_label)

        root_label = s.root(4)
        assert np.dtype(label_type) == type(root_label)



import pytest

import numpy as np
import blobid.plic.analytic_relations as analytic_relations

COUNT = 20000


@pytest.fixture
def slopes() -> np.ndarray:
    """
    List of slopes where 0 <= m1 <= m2 <= m2 < 1
    """
    m = np.random.rand(COUNT, 3)

    m[:, 1] = m[:, 1] * m[:, 2]
    m[:, 0] = m[:, 0] * m[:, 1]
    m = m / np.sum(m, 1, keepdims=True)

    return m


@pytest.fixture
def normals() -> np.ndarray:
    """
    List of normals
    """
    return (np.random.rand(COUNT, 3) * 3) - 1.5


def test_round_trip(normals):
    for n1, n2, n3, f in zip(normals[:, 0], normals[:, 1], normals[:, 2], np.random.rand(COUNT)*0.5):
        alpha = analytic_relations.inverse_problem(n1, n2, n3, f)
        f_test = analytic_relations.forward_problem(n1, n2, n3, alpha)

        assert f_test == pytest.approx(f)


def test_scaled_round_trip_forward(slopes):
    for m1, m2, m3, V in zip(slopes[:, 0], slopes[:, 1], slopes[:, 2], np.random.rand(COUNT)*0.5):
        alpha = analytic_relations._inverse_problem_scaled(m1, m2, m3, V)
        V_test = analytic_relations._forward_problem_scaled(m1, m2, m3, alpha)

        assert V_test == pytest.approx(V)


def test_scaled_round_trip_backward(slopes):
    for m1, m2, m3, alpha in zip(slopes[:, 0], slopes[:, 1], slopes[:, 2], np.random.rand(COUNT)*0.5):
        V = analytic_relations._forward_problem_scaled(m1, m2, m3, alpha)
        alpha_test = analytic_relations._inverse_problem_scaled(m1, m2, m3, V)

        assert alpha_test == pytest.approx(alpha)


def test_scaled_empty(slopes):
    for m1, m2, m3 in zip(slopes[:, 0], slopes[:, 1], slopes[:, 2]):
        assert analytic_relations._forward_problem_scaled(m1, m2, m3, 0.0) == 0
        assert analytic_relations._inverse_problem_scaled(m1, m2, m3, 0.0) == 0


def test_scaled_half_full(slopes):
    for m1, m2, m3 in zip(slopes[:, 0], slopes[:, 1], slopes[:, 2]):
        assert analytic_relations._forward_problem_scaled(m1, m2, m3, 0.5) == pytest.approx(0.5)
        assert analytic_relations._inverse_problem_scaled(m1, m2, m3, 0.5) == pytest.approx(0.5)



"""
Implementation of No Inversion VOF Interface Reconstruction Algorithm (NIVIRA)

Reference
---------

G. D. Weymouth & D. K.-P. Yue, "Conservative Volume-of-Fluid method for free-surface simulations on
Cartesian-grids," Journal of Computational Physics, vol. 229, pp. 2853--2865, 2010.
[10.1016/j.jcp.2009.12.018](https://doi.org/10.1016/j.jcp.2009.12.018)
"""
import warnings
import numpy as np
import numpy.typing as npt

from . import central_difference

from .._numba_support import njit, numba_availible


def get_normals(f: np.ndarray, normals_dtype: npt.DTypeLike) -> np.ndarray:
    """
    Calculate the sign of the interface normal using NIVIRA
    """

    # Numba only supports float32 and float64
    if numba_availible and (f.dtype not in (np.float32, np.float64)):
        warnings.warn(f"void_fraction type {f.dtype.name} not supported, using float32")
        f = f.astype(np.float32)

    if numba_availible and (np.dtype(normals_dtype) not in (np.float32, np.float64)):
        warnings.warn(f"normals_dtype {np.dtype(normals_dtype).name} not supported, using float32")
        normals_dtype = np.float32

    # estimate dominant direction
    d_dom = _get_dominant_direction(f)

    # calculate normals using 3x3x3 stencil described by Weymouth and Yue
    n = np.empty((3, f.shape[0]-2, f.shape[1]-2, f.shape[2]-2), dtype=normals_dtype)

    for d in range(3):
        n[d] = _normals_WY(f, d_dom, d, normals_dtype)

    return n


def _get_dominant_direction(f):
    # estimate the normal using central difference
    n_approx = central_difference.get_normals(f, normals_dtype=f.dtype)

    # cells do not have a normal if empty, full, or undefined CD
    invalid = (f[1:-1, 1:-1, 1:-1] == 0) | (f[1:-1, 1:-1, 1:-1] == 1) | np.all(n_approx == 0, axis=0)

    # calculate the dominant direction, use d_dom=-1 to indicate no valid normal
    d_dom = np.where(invalid, -1, np.argmax(np.abs(n_approx), axis=0))

    return d_dom


@njit
def _normals_WY(f, d_dom, d, dtype) -> np.ndarray:
    n = np.zeros(d_dom.shape, dtype=dtype)

    for i, j, k in np.argwhere(d_dom != -1):
        if d_dom[i, j, k] == d:
            # extract column in direction d
            block = np.swapaxes(f[i:i+3, j:j+3, k:k+3], 0, d)

            # use central difference
            n[i, j, k] = (block[0, 1, 1] - block[2, 1, 1])/2.0

        else:
            # skip the dimension we don't care about
            d_drop = 3 - d - d_dom[i, j, k]
            block = np.take(f[i:i+3, j:j+3, k:k+3], indices=1, axis=d_drop)

            # sum over d_dom
            height = np.sum(block, axis=1 if (d_dom[i, j, k] > d) else 0)

            # central difference of summed columns
            n_tmp = (height[0] - height[2])/2.0

            if abs(n_tmp) < 0.5:
                n[i, j, k] = n_tmp
            else:
                # for steep interface, use one-sided differences
                if n_tmp * f[i+1, j+1, k+1] >= 0:
                    n[i, j, k] = (height[1] - height[2])/1.0
                else:
                    n[i, j, k] = (height[0] - height[1])/1.0

    return n



r"""
Implements the analytical relationships described by Scardovelli and Zaleski[1]_.

.. [1] R. Scardovelli & S. Zaleski, "Analytical Relations Connecting Linear Interfaces and Volume Fractions in
   Rectangular Grids," Journal of Computational Physics, vol. 164, pp. 228-237, 2000.
   [10.1006/jcph.2000.6567](https://doi.org/10.1006/jcph.2000.6567)
"""
import numpy as np


def forward_problem(n1, n2, n3, alpha):
    r"""
    Solve the three-dimensional forward problem, :math:`f = f(\vec{n}, \alpha)`
    """

    # scale the slope and sort in ascending order
    t = abs(n1) + abs(n2) + abs(n3)
    m1, m2, m3 = sorted([abs(n1)/t, abs(n2)/t, abs(n3)/t])

    # rotate so all normals are positive
    if n1 < 0:
        alpha = alpha - n1
    if n2 < 0:
        alpha = alpha - n2
    if n3 < 0:
        alpha = alpha - n3

    # scale alpha
    alpha = alpha / t

    # Interface is outside the cell
    if alpha >= 1:
        return 1
    if alpha <= 0:
        return 0

    # Interface is inside the cell
    if alpha > 0.5:
        return 1.0 - _forward_problem_scaled(m1, m2, m3, 1-alpha)
    return _forward_problem_scaled(m1, m2, m3, alpha)


def _forward_problem_scaled(m1, m2, m3, alpha):
    assert 0 <= alpha <= 0.5
    assert 0 <= m1 <= m2 <= m3

    m12 = m1 + m2
    p = 6*m1*m2*m3

    V1 = (1 if m2 == 0 else m1/m2) * (m1/(6*m3))  # this form avoids the need for approximation

    if alpha < m1:
        return alpha**3 / p
    if alpha < m2:
        return alpha*(alpha-m1) / (2*m2*m3) + V1
    if alpha < min(m12, m3):
        return (alpha**2 * (3*m12 - alpha) + m1**2 * (m1 - 3*alpha) + m2**2 * (m2 - 3*alpha))/p
    if m3 < m12:
        return (alpha**2 * (3 - 2*alpha) + m1**2 * (m1 - 3*alpha) + m2**2 * (m2 - 3*alpha) + m3**2 * (m3 - 3*alpha))/p

    return (2*alpha - m12)/(2*m3)


def inverse_problem(n1, n2, n3, f):
    r"""
    Solve the three-dimensional inverse problem, :math:`\alpha = \alpha(\vec{n}, f)`

    Only valid for :math:`f\in (0, 1)`

    """
    assert 0 < f < 1.0

    # scale the slope and sort in ascending order
    t = abs(n1) + abs(n2) + abs(n3)
    m1, m2, m3 = sorted([abs(n1)/t, abs(n2)/t, abs(n3)/t])

    # calculate the scaled alpha
    if f > 0.5:
        alpha = 1 - _inverse_problem_scaled(m1, m2, m3, 1 - f)
    else:
        alpha = _inverse_problem_scaled(m1, m2, m3, f)

    # unscale alpha
    alpha = alpha * t

    # rotate to original orientation
    if n1 < 0:
        alpha = alpha + n1
    if n2 < 0:
        alpha = alpha + n2
    if n3 < 0:
        alpha = alpha + n3

    return alpha


def _inverse_problem_scaled(m1, m2, m3, V):
    r"""
    Solve the three-dimensional inverse problem
    """

    def cubic_root(c3, c2, c1, c0):
        a0 = c0/c3
        a1 = c1/c3
        a2 = c2/c3

        p = a1/3 - a2**2/9
        q = (a1*a2 - 3*a0)/6 - a2**3/27

        theta = np.arccos(q/np.sqrt(-p**3))/3

        return np.sqrt(-p) * (np.sqrt(3)*np.sin(theta) - np.cos(theta)) - a2/3

    assert 0 <= V <= 0.5
    assert 0 <= m1 <= m2 <= m3

    m12 = m1 + m2
    p = 6*m1*m2*m3

    V1 = (1 if m2 == 0 else m1/m2) * (m1/(6*m3))  # this form avoids the need for approximation
    V2 = V1 + (m2 - m1)/(2 * m3)
    V3 = (m3**2 * (3 * m12-m3) + m1**2 * (m1 - 3 * m3) + m2**2 * (m2 - 3 * m3)) / p if m3 < m12 else 0.5*m12/m3

    if V < V1:
        return (p*V)**(1/3)
    if V < V2:
        return 0.5 * (m1+np.sqrt(m1**2 + 8 * m2 * m3 * (V-V1)))
    if V < V3:
        return cubic_root(
            -1, 3*m12, -3*(m1**2 + m2**2), m1**3 + m2**3 - p*V
        )
    if m3 < m12:
        return cubic_root(
            -2, 3, -3*(m1**2 + m2**2 + m3**2), m1**3 + m2**3 + m3**3 - p*V
        )

    return m3*V + m12/2



"""
Calculate the sign of the interface normal based on a 3x3(x3) stencil
"""
import numpy as np
import numpy.typing as npt

from . import analytic_relations

from .central_difference import get_normals as get_normals_CD
from .nivira import get_normals as get_normals_WY


SUPPORTED_METHODS = ['CD', 'WY']
"""Supported methods for calculating interface normals"""


def get_normals(
        void_fraction: np.ndarray,
        normals_method: str = 'CD',
        normals_type: npt.DTypeLike | None = None
) -> np.ndarray:
    r"""
    Calculate the the interface normals :math:`\vec{n}=\vec{n}(f)` using the method specified by `normals_method`

    Parameters
    ----------
    void_fraction : ndarray[ni+2, nj+2, nk+2]
        The void fractions :math:`f` for each grid cell.
    normals_method: {'CD', 'WY'}, optional
        Method used to calculate interface normals. Defaults to 'CD'.
    normals_type: DTypeLike
        Determines the data-type of `normals`. Defaults to same type as `void_fraction`

    Returns
    -------
        normals : ndarray[3, ni, nj, nk]
            `normals[:,i,j,k]` contains :math:`\vec{n}=\langle n_x, n_y, n_z\rangle` for cell `[i,j,k]`.
            :math:`\vec{n}=\langle 0, 0, 0\rangle` indicates the interface in the cell is undefined.

    Methods
    -------
    - If `normals_method` is `CD`, central differencing is used. For example
    .. math::
        n_x = - \frac{f_{i+1} - f_{i-1}}{2}

    - If `normals_method` is `WY`, normals are calculated using the No Inversion VOF Interface Reconstruction Algorithm
    (NIVIRA) described by Weymouth and Yue.[1]_ `WY` is more accurate than `CD` but can be slower
    (especially if not using Numba).

    For simplicity, all methods assume constant grid spacing.


    .. [1] G. D. Weymouth & D. K.-P. Yue, "Conservative Volume-of-Fluid method for free-surface simulations on
       Cartesian-grids," Journal of Computational Physics, vol. 229, pp. 2853--2865, 2010.
       [10.1016/j.jcp.2009.12.018](https://doi.org/10.1016/j.jcp.2009.12.018)

    """
    if normals_type is None:
        normals_type = void_fraction.dtype

    # checks
    assert void_fraction.ndim == 3
    assert all(dim > 2 for dim in void_fraction.shape)

    match normals_method:
        case 'CD':
            return get_normals_CD(void_fraction, normals_type)
        case 'WY':
            return get_normals_WY(void_fraction, normals_type)
        case _:
            raise ValueError(f"normals_method '{normals_method}' is not supported")


def get_alpha(
        void_fraction: np.ndarray,
        normals: np.ndarray
) -> np.ndarray:
    r"""
    Calculate the the interface intercept :math:`\alpha=\alpha(f, \vec{n})` using
    `blobid.plic.analytic_relations`

    Parameters
    ----------
    void_fraction : ndarray[ni, nj, nk]
        The void fractions :math:`f` for each grid cell.
    normals : ndarray[3, ni, nj, nk]
        The interface normals for each cell.

    Returns
    -------
        alpha : ndarray[ni, nj, nk]
            `alpha[i,j,k]` contains :math:`\alpha` for cell `[i,j,k]`.
            If there is no interface or `normals` is undefined in the cell, :math:`\alpha=\text{NaN}`.
    """

    with np.nditer([void_fraction, normals[0, :, :, :], normals[1, :, :, :], normals[2, :, :, :], None],
                   flags=['common_dtype'], op_dtypes=void_fraction.dtype) as itr:
        for f, n1, n2, n3, a in itr:
            if f <= 0 or f >= 1 or (n1 == 0 and n2 == 0 and n3 == 0):
                # alpha is undefined
                a[...] = np.nan
            else:
                a[...] = analytic_relations.inverse_problem(n1, n2, n3, f)

        return itr.operands[-1]


def get_void_fraction(
        alpha: np.ndarray,
        normals: np.ndarray
) -> np.ndarray:
    r"""
    Calculate the the void fraction :math:`f=f(\alpha, \vec{n})` using
    `blobid.plic.analytic_relations`

    Parameters
    ----------
    alpha : ndarray[ni, nj, nk]
        The interface intercept :math:`\alpha` for each grid cell.
    normals : ndarray[3, ni, nj, nk]
        The interface normals for each cell.

    Returns
    -------
        void_fraction : ndarray[ni, nj, nk]
            `void_fraction[i,j,k]` contains :math:`f` for cell `[i,j,k]`.
            If `normals` is undefined in the cell, :math:`f=\text{NaN}`.
    """

    with np.nditer([alpha, normals[0, :, :, :], normals[1, :, :, :], normals[2, :, :, :], None],
                   flags=['common_dtype'], op_dtypes=alpha.dtype) as itr:
        for a, n1, n2, n3, f in itr:
            if np.isnan(a) or (n1 == 0 and n2 == 0 and n3 == 0):
                # void fraction is undefined
                f[...] = np.nan
            else:
                f[...] = analytic_relations.forward_problem(n1, n2, n3, a)

        return itr.operands[-1]



"""
Tools for keeping track of blob labels
"""
from collections import namedtuple

import numpy as np

from .._numba_support import njit

_DatabaseStorage = namedtuple('_DatabaseStorage', 'r, n, t')
_END_OF_SET = 0


def _root(data: _DatabaseStorage, x) -> np.uint32 | np.integer:
    return data.r[x]


@njit
def _merge(data: _DatabaseStorage, x, y) -> np.uint32 | np.integer:
    # root of each set
    x_root = data.r[x]
    y_root = data.r[y]

    # do nothing if they have the same roots
    if x_root == y_root:
        return x_root

    # combine the larger labeled set into the smaller one
    if x_root < y_root:
        u = x_root
        v = y_root
    else:
        u = y_root
        v = x_root

    i = v
    while i != _END_OF_SET:
        data.r[i] = u
        i = data.n[i]

    data.n[data.t[u]] = v
    data.t[u] = data.t[v]

    return u


class LabelDatabase:
    """
    Keeps track of which labels have been assigned and can combine the labels into connected sets.

    Reference: [He, Chao and Suzuki 2007.
    *A Linear-Time Two-Scan Labeling Algorithm*](https://doi.org/10.1109/ICIP.2007.4379810).
    """

    def __init__(self, label_count, label_type: type = np.uint32):
        """Create an empty LabelDatabase"""

        assert np.issubdtype(label_type, np.integer), 'label type must be an integer'
        self.label_type = label_type
        """The type used for labels, set by `__init__`"""

        # make sure we wont overflow
        if label_count > np.iinfo(self.label_type).max:
            raise OverflowError(
                "Number of labels exceeds capacity of label_type " + self.label_type.__name__)

        # initialize the database with unconnected labels up to label_count
        self.data = _DatabaseStorage(
            np.arange(label_count+1, dtype=self.label_type),
            np.full(label_count+1, _END_OF_SET, dtype=self.label_type),
            np.arange(label_count+1, dtype=self.label_type)
        )

    def get_labels(self) -> np.ndarray:
        """Return a list of all the labels in the database"""
        return np.arange(1, len(self.data.r), dtype=self.label_type)

    def merge(self, x, y) -> np.integer:
        """
        Join the set containing label `x` and the set containing label `y`

        Returns: the root of the combined set

        """
        return _merge(self.data, x, y)

    def root(self, x) -> np.uint32 | np.integer:
        """Returns the smallest label in the set containing label `x`"""
        return _root(self.data, x)

    def get_sequential_lookup_table(self) -> np.ndarray:
        """
        Get a lookup table for consecutive final labels
        """

        lookup = np.empty_like(self.data.r)

        # for roots, r[i] == i
        root_indices = np.flatnonzero(self.data.r == np.arange(len(self.data.r)))

        # label roots with new consecutive labels
        lookup[root_indices] = np.arange(len(root_indices))

        # For all labels (root and non-root), their root is r[i].
        lookup = lookup[self.data.r]

        return lookup
